/*
Sand3d
Copyright (C) 2007 - David Pascoe-Deslauriers

  This file is part of Sand3d.

    Sand3d is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Sand3d is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sand3d; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
    
This particle system is based on an example from http://nehe.gamedev.net/ which was free code, it has been heavily modified.
*/
#include <stdlib.h>
#include "fountain.h"

fountain::fountain(){
    for (int i=0;i<MAX_PARTICLES;i++){				// Initials All The Textures
		particle[i].active=true;								// Make All The Particles Active
		particle[i].life=1.0f;								// Give All The Particles Full Life
		particle[i].fade=float(rand()%100)/1000.0f+0.3f;	// Random Fade Speed
		particle[i].r=1.0;
		particle[i].g=0.3;
		particle[i].b=0.0;
		particle[i].xi=float((rand()%50)-26.0f)*10.0f;		// Random Speed On X Axis
		particle[i].yi=float((rand()%50)-25.0f)*10.0f;		// Random Speed On Y Axis
		particle[i].zi=float((rand()%50)-25.0f)*10.0f;		// Random Speed On Z Axis
		particle[i].xg=0.0f;
		particle[i].yg=1.0;
		particle[i].zg=0.0f;
		particle[i].x = 0;
		particle[i].y = 0;
		particle[i].z = 0;
	}
}

fountain::~fountain(){}

void fountain::draw(GLuint tex, int x_center, int y_center, int z_center){

	glClearDepth(1.0f);									// Depth Buffer Setup
	glDisable(GL_DEPTH_TEST);							// Disable Depth Testing

	glBlendFunc(GL_SRC_ALPHA,GL_ONE);					// Type Of Blending To Perform
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D,tex);
	
	for (int i=0;i<MAX_PARTICLES;i++) {
		if (particle[i].active)	{
			float x=particle[i].x;						// Grab Our Particle X Position
			float y=particle[i].y;						// Grab Our Particle Y Position
			float z=particle[i].z;					// Particle Z Pos + Zoom

			// Draw The Particle Using Our RGB Values, Fade The Particle Based On It's Life
			glColor4f(particle[i].r,particle[i].g,particle[i].b,particle[i].life);
			glPushMatrix();

			GLfloat matrix[16];
			glGetFloatv(GL_MODELVIEW_MATRIX,matrix);
			matrix[0] = 1.0f; matrix[1] = 0.0f; matrix[2] = 0.0f;
			matrix[4] = 0.0f; matrix[5] = 1.0f; matrix[6] = 0.0f;
			glLoadMatrixf(matrix);

			glBegin(GL_TRIANGLE_STRIP);						// Build Quad From A Triangle Strip
			    glTexCoord2d(1,1); glVertex3f(x+1.5,y+1.5,z); // Top Right
				glTexCoord2d(0,1); glVertex3f(x-1.5,y+1.5,z); // Top Left
				glTexCoord2d(1,0); glVertex3f(x+1.5,y-1.5,z); // Bottom Right
				glTexCoord2d(0,0); glVertex3f(x-1.5,y-1.5,z); // Bottom Left
			glEnd();										// Done Building Triangle Strip
			glPopMatrix();
		}
    }

  	glDisable(GL_TEXTURE_2D);
  	glEnable(GL_DEPTH_TEST);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);
}
    
void fountain::step(){
	for (int i=0;i<MAX_PARTICLES;i++) {
		if (particle[i].active) {
			float x=particle[i].x;						// Grab Our Particle X Position
			float y=particle[i].y;						// Grab Our Particle Y Position
			float z=particle[i].z;					// Particle Z Pos + Zoom

			particle[i].x+=particle[i].xi/(slowdown*500);// Move On The X Axis By X Speed
			particle[i].y+=particle[i].yi/(slowdown*500);// Move On The Y Axis By Y Speed
			particle[i].z+=particle[i].zi/(slowdown*500);// Move On The Z Axis By Z Speed

			particle[i].xi+=particle[i].xg;			// Take Pull On X Axis Into Account
			particle[i].yi+=particle[i].yg;			// Take Pull On Y Axis Into Account
			particle[i].zi+=particle[i].zg;			// Take Pull On Z Axis Into Account
			particle[i].life-=particle[i].fade;		// Reduce Particles Life By 'Fade'

			if (particle[i].life<0.0f) {
				particle[i].life=1.0f;					// Give It New Life
				particle[i].fade=float(rand()%100)/2000.0f+0.008f;	// Random Fade Value
				particle[i].x=(rand()%100-50)/200;
				particle[i].y=(rand()%100-50)/200;
				particle[i].z=(rand()%100-50)/200;
				particle[i].xi=float((rand()%40)-20.0f);	// X Axis Speed And Direction
				particle[i].yi=float((rand()%60)-20.0f);	// Y Axis Speed And Direction
				particle[i].zi=float((rand()%40)-20.0f);	// Z Axis Speed And Direction
			}
		}
    } 
}
	
