#include "loadTextures.h"

/*
The loadBMP and loadGLTexture functions are based on examples from http://nehe.gamedev.net/.  These functions are free code.
*/

/* simple loader for 24bit bitmaps (data is in rgb-format) */
int loadBMP(char *filename, textureImage *texture){
    FILE *file;
    unsigned short int bfType;
    long int bfOffBits;
    short int biPlanes;
    short int biBitCount;
    long int biSizeImage;
    int i;
    unsigned char temp;
    /* make sure the file is there and open it read-only (binary) */
    if ((file = fopen(filename, "rb")) == NULL){
        return 0;
    }
    if(!fread(&bfType, sizeof(short int), 1, file)){
        return 0;
    }
    /* check if file is a bitmap */
    if (bfType != 19778){
        return 0;
    }        
    /* get the file size */
    /* skip file size and reserved fields of bitmap file header */
    fseek(file, 8, SEEK_CUR);
    /* get the position of the actual bitmap data */
    if (!fread(&bfOffBits, sizeof(long int), 1, file)){
        return 0;
    }
    /* skip size of bitmap info header */
    fseek(file, 4, SEEK_CUR);
    /* get the width of the bitmap */
    fread(&texture->width, sizeof(int), 1, file);
    /* get the height of the bitmap */
    fread(&texture->height, sizeof(int), 1, file);
    /* get the number of planes (must be set to 1) */
    fread(&biPlanes, sizeof(short int), 1, file);
    if (biPlanes != 1){
        return 0;
    }
    /* get the number of bits per pixel */
    if (!fread(&biBitCount, sizeof(short int), 1, file)){
        return 0;
    }
    if (biBitCount != 24){
        return 0;
    }
    /* calculate the size of the image in bytes */
    biSizeImage = texture->width * texture->height * 3;
    texture->data = (unsigned char*)malloc(biSizeImage);
    /* seek to the actual data */
    fseek(file, bfOffBits, SEEK_SET);
    if (!fread(texture->data, biSizeImage, 1, file)){
        return 0;
    }
    /* swap red and blue (bgr -> rgb) */
    for (i = 0; i < biSizeImage; i += 3){
        temp = texture->data[i];
        texture->data[i] = texture->data[i + 2];
        texture->data[i + 2] = temp;
    } 
    return 1;
}

int loadGLTextures(GLuint *tex) {
        int Status=false;
        textureImage *texti;				// Create Storage Space For The Textures
        texti = (textureImage*)malloc(sizeof(textureImage));
        if (loadBMP("particle.bmp", texti)){
			Status=true;
			glGenTextures(1, &tex[0]);				// Create One Texture

			glBindTexture(GL_TEXTURE_2D, tex[0]);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0, GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        }

        if (texti){
			if (texti->data){
				free(texti->data);			// Free The Texture Image Memory
			}
			free(texti);						// Free The Image Structure
		}
		
		if (loadBMP("particle2.bmp", texti)){
			Status=true;
			glGenTextures(1, &tex[1]);				// Create One Texture

			glBindTexture(GL_TEXTURE_2D, tex[1]);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0, GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        }

        if (texti){
			if (texti->data){
				free(texti->data);			// Free The Texture Image Memory
			}
			free(texti);						// Free The Image Structure
		}
		
				if (loadBMP("particle3.bmp", texti)){
			Status=true;
			glGenTextures(1, &tex[2]);				// Create One Texture

			glBindTexture(GL_TEXTURE_2D, tex[2]);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexImage2D(GL_TEXTURE_2D, 0, 3, texti->width, texti->height, 0, GL_RGB, GL_UNSIGNED_BYTE, texti->data);
        }

        if (texti){
			if (texti->data){
				free(texti->data);			// Free The Texture Image Memory
			}
			free(texti);						// Free The Image Structure
		}
        return Status;									// Return The Status
}
