/*
Sand3d
Copyright (C) 2007 - David Pascoe-Deslauriers

  This file is part of Sand3d.

    Sand3d is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Sand3d is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Sand3d; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "world.h"
#include <stdlib.h>
#include <iostream>

world::world(int a, int b, int c){
    x = a;
    y = b;
    z = c;
    w = new particle**[x];
    for(int i = 0; i < x; i++){
        w[i] = new particle*[y];
        for(int j = 0; j < y; j++){
                w[i][j] = new particle[z];
        }
    }
    
    
    types = new particle[ptypes];
    for(int i = 0; i < ptypes; i++){
        types[i].force = false;
        types[i].fire = -1;
        types[i].secondary_counter = -1;
        types[i].explodes = -1;
            switch (i){
            case EMPTY:
                sprintf(types[i].name,"Eraser");
                types[i].mass = 0;
                types[i].type = i;
                types[i].colour.x = 0;
                types[i].colour.y = 0;
                types[i].colour.z = 0;
                types[i].alpha = 0;
                types[i].combustion_counter = -1;
                types[i].phase = SKYHOOK;
                types[i].fire = -1;
                types[i].puts_out_fire = false;
                break;
            case OIL:
                sprintf(types[i].name,"Oil");
                types[i].mass = 2;
                types[i].type = i;
                types[i].colour.x = 0.2;
                types[i].colour.y = 0.1;
                types[i].colour.z = 0.01;
                types[i].alpha = 0.9;
                types[i].combustion_counter = 150;               
                types[i].phase = LIQUID;
                types[i].fire = 10;
                types[i].puts_out_fire = false;
                break;
            case WATER:
                sprintf(types[i].name,"Water");
                types[i].mass = 5;
                types[i].type = i;
                types[i].colour.x = 0.1;
                types[i].colour.y = 0.1;
                types[i].colour.z = 1;
                types[i].alpha = 0.5;
                types[i].phase = LIQUID;
                types[i].combustion_counter = -1; 
                types[i].puts_out_fire = true;                            
                break;
            case SAND:
                sprintf(types[i].name,"Sand");
                types[i].mass = 10;
                types[i].type = i;
                types[i].colour.x = 1;
                types[i].colour.y = 0.82;
                types[i].colour.z = 0.38;
                types[i].alpha = 1.0f;
                types[i].combustion_counter = -1;
                types[i].phase = SOLID;
                types[i].puts_out_fire = false;
                break;         
            case FIRE:
                sprintf(types[i].name,"Fire");
                types[i].mass = 0.01;
                types[i].type = i;
                types[i].colour.x = 1.0;
                types[i].colour.y = 0.24;
                types[i].colour.z = 0.02;
                types[i].alpha = 0.6;
                types[i].combustion_counter = 50;
                types[i].fire = 0;
                types[i].phase = SOLID;
                types[i].puts_out_fire = false;
                break;
            case GLASS:
                sprintf(types[i].name,"Glass");
                types[i].mass = 100;
                types[i].type = i;
                types[i].colour.x = 0.3;
                types[i].colour.y = 0.8;
                types[i].colour.z = 1.0;
                types[i].alpha = 0.12f;
                types[i].phase = HOLLOW;
                types[i].combustion_counter = -1;
                types[i].puts_out_fire = false;
                break;
            case WALL:
                sprintf(types[i].name,"Wall");
                types[i].mass = 100;
                types[i].type = i;
                types[i].colour.x = 0.5;
                types[i].colour.y = 0.5;
                types[i].colour.z = 0.5;
                types[i].alpha = 1.0f;
                types[i].phase = HOLLOW;
                types[i].combustion_counter = -1;
                types[i].puts_out_fire = false;
                break;
            case METHANE:
                sprintf(types[i].name,"Methane");
                types[i].mass = -1;
                types[i].type = i;
                types[i].colour.x = 0.9;
                types[i].colour.y = 0.9;
                types[i].colour.z = 0.9;
                types[i].alpha = 0.2f;
                types[i].phase = GAS;
                types[i].combustion_counter = 10;
                types[i].fire = 1;
                types[i].puts_out_fire = false;
                break;
            case H2S:
                sprintf(types[i].name,"Hydrogen sulfide");
                types[i].mass = 0.5;
                types[i].type = i;
                types[i].colour.x = 0.9;
                types[i].colour.y = 0.9;
                types[i].colour.z = 0.0;
                types[i].alpha = 0.2f;
                types[i].phase = GAS;
                types[i].combustion_counter = 20;
                types[i].fire = 5;
                types[i].puts_out_fire = false;
                break;
            case CO2:
                sprintf(types[i].name,"Carbon Dioxide");
                types[i].mass = 0.25;
                types[i].type = i;
                types[i].colour.x = 0.2;
                types[i].colour.y = 0.2;
                types[i].colour.z = 0.2;
                types[i].alpha = 0.2f;
                types[i].phase = GAS;
                types[i].combustion_counter = -1;
                types[i].fire = -1;
                types[i].puts_out_fire = true;
                break;
            case N2O:
                sprintf(types[i].name,"Nitrogen Dioxide");
                types[i].mass = 0;
                types[i].type = i;
                types[i].colour.x = 0.6;
                types[i].colour.y = 0.3;
                types[i].colour.z = 0;
                types[i].alpha = 0.2f;
                types[i].phase = GAS;
                types[i].combustion_counter = -1;
                types[i].fire = -1;
                types[i].puts_out_fire = false;
                break;
            case TORCH:
                sprintf(types[i].name,"Torch");
                types[i].mass = 0.01;
                types[i].type = i;
                types[i].colour.x = 1.0;
                types[i].colour.y = 0.24;
                types[i].colour.z = 0.02;
                types[i].alpha = 0.6;
                types[i].combustion_counter = 500;
                types[i].fire = 0;
                types[i].phase = HOLLOW;
                types[i].puts_out_fire = false;
                break;
            case SPOUT:
                sprintf(types[i].name,"Spout");
                types[i].mass = 100;
                types[i].type = i;
                types[i].colour.x = 0.3;
                types[i].colour.y = 0.3;
                types[i].colour.z = 1;
                types[i].alpha = 1;
                types[i].combustion_counter = -1;
                types[i].fire = -1;
                types[i].phase = GENERATOR;
                types[i].puts_out_fire = false;
                types[i].secondary_counter = WATER;
                break;
            case WELL:
                sprintf(types[i].name,"Well");
                types[i].mass = 100;
                types[i].type = i;
                types[i].colour.x = 0.3;
                types[i].colour.y = 0.3;
                types[i].colour.z = 0;
                types[i].alpha = 1;
                types[i].combustion_counter = 500;
                types[i].fire = 50;
                types[i].phase = GENERATOR;
                types[i].puts_out_fire = false;
                types[i].secondary_counter = OIL;
                break; 
            case COW:
                sprintf(types[i].name,"Cow");
                types[i].mass = 100;
                types[i].type = i;
                types[i].colour.x = 1;
                types[i].colour.y = 1;
                types[i].colour.z = 1;
                types[i].alpha = 1;
                types[i].combustion_counter = 30;
                types[i].fire = 50;
                types[i].phase = GENERATOR;
                types[i].puts_out_fire = false;
                types[i].secondary_counter = METHANE;
                break;                
        }
    }
    for (int a = 0; a < x; a++){
        for (int b = 0; b < y; b++){
            for(int c = 0; c < z; c++){
                add_particle(EMPTY,a,b,c);   
            }
        }
    }
}

world::~world(){
    for(int i = 0; i < x; i++){
        for(int j = 0; j < y; j++){
            delete[] w[i][j];
        }
        delete[] w[i];
    }
    delete[] w;
}    
        
int world::add_particle(int type,int a,int b,int c){
    if (a < 0) a = 0;
    if (b < 0) b = 0;
    if (c < 0) c = 0;
    if (a > x-1) a = x-1;
    if (b > y-1) b = y-1;
    if (c > z-1) c = z-1;
    
    w[a][b][c] = types[type];
}    

void world::motion(int a,int b,int c){
    if(b>0 && w[a][b][c].mass > w[a][b-1][c].mass){
        swap(a,b,c,a,b-1,c);
        return;
    } else {
        int order = 0;
        int rank[8];
        for(int n=0;n<8;n++) rank[n] = 0;
        if (a>0 && b>0 && w[a][b][c].mass > w[a-1][b-1][c].mass){
            rank[order++] = 0;
            //swap(a,b,c,a-1,b-1,c);
        }
        
        if (b>0 && c>0 && w[a][b][c].mass > w[a][b-1][c-1].mass){
            rank[order++] = 1;            
            //swap(a,b,c,a,b-1,c-1);
        }
        
        if (a<x-1 && b>0 && w[a][b][c].mass > w[a+1][b-1][c].mass){
            rank[order++] = 2;
            //swap(a,b,c,a+1,b-1,c);
        }
        
        if (b>0 && c<z-1 && w[a][b][c].mass > w[a][b-1][c+1].mass){
            rank[order++] = 3;
            //swap(a,b,c,a,b-1,c+1);
        }    
        
        if (a>0 && b>0 && c>0&& w[a][b][c].mass > w[a-1][b-1][c-1].mass){
            rank[order++] = 4;
            //swap(a,b,c,a-1,b-1,c-1);       
        }
        
        if (a>0 && b>0 && c<z-1 && w[a][b][c].mass > w[a-1][b-1][c+1].mass){
            rank[order++] = 5;
            //swap(a,b,c,a-1,b-1,c+1);      
        }
        
        if (a<x-1 && b>0 && c<z-1&& w[a][b][c].mass > w[a+1][b-1][c+1].mass){
            rank[order++] = 6;
            //swap(a,b,c,a+1,b-1,c+1);    
        }
        
        if (a<x-1 && b>0 && c>0&& w[a][b][c].mass > w[a+1][b-1][c-1].mass){
            rank[order++] = 7;
            //swap(a,b,c,a+1,b-1,c-1);   
        }

        if(order > 0){
            w[a][b][c].force = false;
            int r = rand()%order;
            switch (rank[r]) {
                case 0:
                    swap(a,b,c,a-1,b-1,c);
                break;
                case 1:
                    swap(a,b,c,a,b-1,c-1);
                break;
                case 2:
                    swap(a,b,c,a+1,b-1,c);
                break;
                case 3:
                    swap(a,b,c,a,b-1,c+1);
                break;
                case 4:
                    swap(a,b,c,a-1,b-1,c-1);
                break;
                case 5:
                    swap(a,b,c,a-1,b-1,c+1);
                break;
                case 6:
                    swap(a,b,c,a+1,b-1,c+1);
                break;
                case 7:
                    swap(a,b,c,a+1,b-1,c-1);
                break;            
            }
            return;
        }

        order = 0;
        int rankx[8];
        int rankz[8];
        
        if (w[a][b][c].phase == LIQUID && b < y){           
            for (int i = -1; i < 2; i++){
                for (int j = -1; j < 2; j++){
                    if (i == 0 && j == 0);
                    else if (a+i >= 0 && a+i < x && c+j >= 0 && c+j < z){
                         if (w[a+i][b+1][c+j].phase == LIQUID && w[a+i][b+1][c+j].mass >= w[a][b][c].mass) {
                             w[a][b][c].force = true;
                         }
                         if ((w[a+i][b][c+j].phase == LIQUID || w[a+i][b][c+j].type == EMPTY)&& w[a+i][b][c+j].mass < w[a][b][c].mass){
                             rankx[order]=a+i;
                             rankz[order]=c+j;
                             order++;
                         }
                    }
                }
            }
        }
        if (order > 0 && w[a][b][c].force == true){
           int pick = rand()%order;
           w[a][b][c].force = false;
           swap(a,b,c,rankx[pick],b,rankz[pick]);   
        }
        if (order <= 0 && w[a][b][c].force == true){
            for (int i = -1; i < 2; i++){
                for (int j = -1; j < 2; j++){
                    if (i == 0 && j == 0);
                    else if (a+i >= 0 && a+i < x && c+j >= 0 && c+j < z){
                         if (w[a+i][b][c+j].phase == LIQUID){
                             w[a+i][b][c+j].force = true;
                         }
                    }
                }
            }
            w[a][b][c].force = false;
        }
                    
    }    
}

void world::gas_motion(int a, int b, int c){
    int order = 0;
    int rankx[26];
    int ranky[26];
    int rankz[26];
    int low = -1;
    int high = 2;
    if (w[a][b][c].mass > 0){
        high = 1;
    } else if (w[a][b][c].mass < 0){
        low = 0;   
    }    
    
    for (int i = -1; i < 2; i++){
        for (int j = -1; j < 2; j++){
            for(int k = low; k < high; k++){
                if (i == 0 && j == 0 && k == 0);
                else if (a+i >= 0 && a+i < x && c+j >= 0 && c+j < z && b+k >= 0 && b+k < y){
                     if (w[a+i][b+k][c+j].type == EMPTY || w[a+i][b+k][c+j].phase == GAS){
                         if  (k == -1 && w[a+i][b+k][c+j].mass > w[a][b][c].mass){
                             rankx[order] = a+i;
                             ranky[order] = b+k;
                             rankz[order] = c+j;
                             ++order;
                         } else if  (k == 1 && w[a+i][b+k][c+j].mass < w[a][b][c].mass){
                             rankx[order] = a+i;
                             ranky[order] = b+k;
                             rankz[order] = c+j;
                             ++order;
                         } else {
                             rankx[order] = a+i;
                             ranky[order] = b+k;
                             rankz[order] = c+j;
                             ++order;
                         }           
                     }
                }
            }    
        }
    }

    if (order > 0){
       int pick = rand()%order;
       swap(a,b,c,rankx[pick],ranky[pick],rankz[pick]);   
    }    
}    

void world::spread_fire(int a,int b,int c){
    for (int i = -1; i <= 1; i++){
          for (int j = -1; j <= 1; j++){
                  for (int k = -1; k <= 1; k++){
                      if (a == 0 && b == 0 && c==0);
                      else if(a+i >= 0 && a+i <= x-1 && b+j >= 0 && b+j <= y-1 && c+k >= 0 && c+k <= z-1){
                          if (w[a][b][c].fire == 0 && w[a+i][b+j][c+k].combustion_counter > 0){
                              if (w[a+i][b+j][c+k].fire > 0){
                                  w[a+i][b+j][c+k].fire--;
                              }
                              if (w[a+i][b+j][c+k].fire == 0){
                                w[a+i][b+j][c+k].colour.x = 1.0;
                                w[a+i][b+j][c+k].colour.y = 0.24;
                                w[a+i][b+j][c+k].colour.z = 0.02;
                                w[a+i][b+j][c+k].alpha = 0.6;
                                
                              }    
                          } else if (j > -1 && w[a+i][b+j][c+k].puts_out_fire){
                              swap(a,b,c,a+i,b+j,c+k);
                              return;
                          }    
                      }    
                  }
          }        
    }    
}

void world::generate(int a, int b, int c){
    if (rand()%3)return;
    int i = rand()%3-1;
    int j = rand()%3-1;
    int k = rand()%3-1;
    if(a+i >= 0 && a+i <= x-1 && b+j >= 0 && b+j <= y-1 && c+k >= 0 && c+k <= z-1){
        if (w[a+i][b+j][c+k].phase != GENERATOR)
            add_particle(w[a][b][c].secondary_counter,a+i,b+j,c+k);
    }    
}    

int world::step(int a,int b,int c){
    if (w[a][b][c].fire == 0){
       w[a][b][c].combustion_counter--; 
       if (w[a][b][c].combustion_counter < 0){
          add_particle(EMPTY,a,b,c);
          return 0;
       }
    }
    if (w[a][b][c].phase == GAS)
        gas_motion(a,b,c);
    else if (w[a][b][c].phase == LIQUID || w[a][b][c].phase == SOLID)
        motion(a,b,c);
    else if (w[a][b][c].phase == GENERATOR)
        generate(a,b,c);
    if (w[a][b][c].fire == 0)
        spread_fire(a,b,c);    
}    

particle world::get_particle(int a,int b,int c){
    return w[a][b][c];    
}

void world::swap(int a, int b, int c, int d, int e, int f){   
    if (w[a][b][c].fire == 0 && w[d][e][f].puts_out_fire){
        add_particle(EMPTY,a,b,c);       
    } else if (w[d][e][f].fire == 0 && w[a][b][c].puts_out_fire){
        add_particle(EMPTY,d,e,f);
    }    
    particle tmp = w[a][b][c];
    w[a][b][c] = w[d][e][f];
    w[d][e][f] = tmp;
    return;   
}    
